/******************************************************************************\

  Copyright 1995 The University of North Carolina at Chapel Hill.
  All Rights Reserved.

  Permission to use, copy, modify and distribute this software and its
  documentation for educational, research and non-profit purposes, without
  fee, and without a written agreement is hereby granted, provided that the
  above copyright notice and the following three paragraphs appear in all
  copies.

  IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE LIABLE
  TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
  DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
  ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF NORTH CAROLINA HAVE BEEN
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.


  Permission to use, copy, modify and distribute this software and its
  documentation for educational, research and non-profit purposes, without
  fee, and without a written agreement is hereby granted, provided that the
  above copyright notice and the following three paragraphs appear in all
  copies.

  THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN
  "AS IS" BASIS, AND THE UNIVERSITY OF NORTH CAROLINA HAS NO OBLIGATION TO
  PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.


   --------------------------------- 
  |Please send all BUG REPORTS to:  |
  |                                 |
  |   geom@cs.unc.edu               |
  |                                 |
   ---------------------------------
  
     
  The authors may be contacted via:

  US Mail:             J. Cohen/M. Lin/D. Manocha/K. Ponamgi
                       Department of Computer Science
                       Sitterson Hall, CB #3175
                       University of N. Carolina
                       Chapel Hill, NC 27599-3175

  Phone:               (919)962-1749

  EMail:               {cohenj,lin,manocha,ponamgi}@cs.unc.edu




\*****************************************************************************/


/*****************************************************************************\
  collision_types.h
  --
  Description : Types for the upper layer of the collision detection library.
                The types for the Lin/Canny layer are in dist.h. 

\*****************************************************************************/

/* Protection from multiple includes. */
#ifndef INCLUDED_COLLISION_TYPES_H
#define INCLUDED_COLLISION_TYPES_H


/*------------------ Includes Needed for Definitions Below ------------------*/

#include <options.h>
#include <dist.h>

/*-------------------------------- Constants --------------------------------*/

#define COL_NULL 0

#define __COL_X 0
#define __COL_Y 1
#define __COL_Z 2

#define __COL_OVERLAP_X    0x1
#define __COL_OVERLAP_Y    0x2
#define __COL_OVERLAP_Z    0x4
#define __COL_OVERLAP_ALL  0x7

/*--------------------------------- Macros ----------------------------------*/


/*  indexing routines -- given the offset from the start of the allocated
    storage, compute the row and column in the half array, or do the reverse */

#ifdef COL_DENSE_MATRIX

#define __COL_PAIR_INDEX(state,row,col) 				\
       ( ((col) < (state).matrix_cols)   				\
	? ( (row)*((state).matrix_cols)+(col) ) 			\
	: (((state).matrix_rows-(row)-1) * ((state).matrix_cols) + 	\
	     ((state).max_polytopes - (col) - 1)) )

#define __COL_PAIR_ROW_COL(state, pair, row, col) 		\
        {              						\
           int offset; 						\
                       						\
           offset = (pair-(state).pairs); 			\
	   (row) = offset / (state).matrix_cols; 		\
           (col) = offset % (state).matrix_cols; 		\
           if ((row) >= (col)) 					\
           {   							\
	       (row) = (state).matrix_rows - (row) - 1; 	\
	       (col) = (state).max_polytopes - (col) - 1; 	\
	   } 							\
        }

#define __COL_GET_PAIR(state,row,col,pair)						\
     {											\
         (pair) = ( ((row) < (col))							\
		  ? (&((state).pairs[__COL_PAIR_INDEX((state), (row), (col))]))		\
		  : (&((state).pairs[__COL_PAIR_INDEX((state), (col), (row))])) );	\
     }

#define __COL_ALLOCATE_PAIR(state,row, col, pair) 	\
          __COL_GET_PAIR((state), (row), (col), (pair))

#define __COL_PAIR_ACTIVE(pair)  (pair)->active



#endif /* COL_DENSE_MATRIX */



#ifdef COL_SPARSE_MATRIX

#define __COL_PAIR_INDEX(state,row,col) 				\
       ( ((col) < (state).matrix_cols)   				\
	? ( (row)*((state).matrix_cols)+(col) ) 			\
	: (((state).matrix_rows-(row)-1) * ((state).matrix_cols) + 	\
	     ((state).max_polytopes - (col) - 1)) )

#define __COL_PAIR_ROW_COL(state, pair, row, col)		\
        {							\
           (row) = (pair)->ids[0];				\
	   (col) = (pair)->ids[1];				\
        }

#define __COL_GET_PAIR(state,row,col,pair)					\
     {										\
         (pair) = ( ((row) < (col))						\
		  ? ((state).pairs[__COL_PAIR_INDEX((state), (row), (col))])	\
		  : ((state).pairs[__COL_PAIR_INDEX((state), (col), (row))]) );	\
     }

#define __COL_ALLOCATE_PAIR(state,row,col,pair)						\
     {											\
	 if ((row) < (col))								\
	 {										\
	     (pair) = (state).pairs[__COL_PAIR_INDEX((state), (row), (col))] =		\
		 (__col_PairNode *)calloc(1, sizeof(__col_PairNode));			\
	     (pair)->ids[0] = (row);							\
	     (pair)->ids[1] = (col);							\
	 }										\
	 else										\
	 {										\
	     (pair) = (state).pairs[__COL_PAIR_INDEX((state), (col), (row))] =		\
		 (__col_PairNode *)calloc(1, sizeof(__col_PairNode));			\
	     (pair)->ids[0] = (col);							\
	     (pair)->ids[1] = (row);							\
	 }										\
	 if (!(pair))									\
	 {										\
	     fprintf(stderr,								\
		     "__COL_ALLOCATE_PAIR: couldn't allocate __col_PairNode\n");	\
	     return COL_ERROR;								\
	 }										\
     }

#define __COL_PAIR_ACTIVE(pair)	    ((!pair) ? COL_FALSE : (pair)->active)


#endif /* COL_SPARSE_MATRIX */




#define __COL_OVERLAPPING(state, id1, id2, overlap)		  	        \
        {								  	\
	    int row, col;						  	\
										\
	    if ((id1)<(id2))							\
	    {									\
		row = (id1);							\
		col = (id2);							\
	    }									\
	    else								\
	    {									\
		row = (id2);							\
		col = (id1);							\
	    }									\
	    (overlap) =							  	\
		((state).overlap_status[__COL_PAIR_INDEX((state),row,col)]	\
		 == __COL_OVERLAP_ALL);					  	\
	}

#define __COL_SWAP_OVERLAP(state, id1, id2, axis)                     	        \
        {									\
	    int row, col;							\
	    									\
	    if ((id1)<(id2))							\
	    {									\
		row = (id1);							\
		col = (id2);							\
	    }									\
	    else								\
	    {									\
		row = (id2);							\
		col = (id1);							\
	    }									\
	    (state).overlap_status[__COL_PAIR_INDEX((state),row,col)] ^=	\
		    (1 << (axis));						\
	}

/*---------------------------------- Types ----------------------------------*/


/*
   A __col_BoxVert is an element of a an extrema list.  Each polytope has 6
   __col_BoxVerts -- xmin, ymin, zmin, xmax, ymax, zmax.
*/
typedef struct __col_box_vert_struct
{
    int           id;        /* number of this vertex's polytope */
    char          min_max;   /* 0 if this is a MIN, 1 if it's a MAX */
    double        value;     /* transformed coordinate (X, Y, or Z) */
    __col_Vertex *vert;      /* Pointer to the actual __col_Vertex */
    
                       /* prev and next __col_BoxVert in this list
			  (either x list, y list, or z list) */
    struct __col_box_vert_struct *prev;
    struct __col_box_vert_struct *next;
} __col_BoxVert;


/*
   Each polytope has a bounding box.  For fixed-sized bounding cubes, a center
   a radius defines the cube.  For dynamic bounding boxes, 6 extremal vertices
   define the box
*/
typedef struct __col_bounding_box_struct
{
    __col_BoxVert min[3];
    __col_BoxVert max[3];
    char use_cuboid;
    double center[3];
    double radius;
} __col_BoundingBox;

/*
   Each instanced object is a __col_Polytope.  It has an id number, a
   polyhedral representation (from dist.h), and a bounding box.
*/
typedef struct col_polytope_struct
{
    int id;
    __col_Polyhedron *polytope;
    __col_BoundingBox box;

} __col_Polytope;


/*
   Every pair of objects that may interact is represented by a pair node (in
   the COL_SPARSE_MATRIX representation, only activated pairs have allocated
   __col_PairNodes).  The pair node stores the current closest features for
   that pair, and the current closest distance.  It also contains flags to
   indicate whether the pair is currently active, which dimensions of the
   pair's bounding boxes are currently overlapping, and whether the pair is
   currently colliding.  In addition, this pair may be inserted and deleted
   from several lists, so the node contains points used for traversing these
   lists.
*/
typedef struct __col_pn_struct
{
#ifdef COL_SPARSE_MATRIX
    int          ids[2];
#endif
    void        *closest_features[2];
    float        distance;
    char         active;                          /* boolean flag */
    char         colliding;                       /* boolean flag */
    struct __col_pn_struct    *next_active;       /* next user-activated pair */
    struct __col_pn_struct    *prev_nbody_active; /* next and previous pairs that have*/
    struct __col_pn_struct    *next_nbody_active; /*   all 3 coordinates overlapping  */
                                                  /*   (i.e. ones requiring Lin/Canny)*/
    struct __col_pn_struct    *next_colliding;    /* next/prev pair in list of */
    struct __col_pn_struct    *prev_colliding;    /*    colliding pairs        */
} __col_PairNode;


/*
   A col_Report is actually a single element of a collision report.  An array
   of these may be past to a report function, which fills in the array.  Each
   report element contains the ids of a pair of colliding polytopes also two
   features which were the closest in the most recent non-colliding test.  One
   might perform further analysis to determine the full set of colliding
   features if necessary.
*/
typedef struct col_report_struct
{
    int     ids[2];
    void   *features[2];  /* __col_Vertex, __col_Edge, or __col_Face -- the
			     tag will tell */
} col_Report;

/*
   A col_DistReport is a single element of a distance report.  An array of
   these may be passed to a distance reporting function, which will report
   known distances (distances are known only for pairs with overlapping
   bounding boxes when the nbody algorithm is enabled or all active pairs
   where the nbody algorithm is disabled.
*/
typedef struct distance_report_struct
{
    int     ids[2];
    void   *features[2]; /* __col_Vertex, __col_Edge, or __col_Face -- the
			     tag will tell */
    float   distance;
} col_DistReport;

/*
   The __col_State structures contains global state for the collision
   detection library.  It contains all the polytope structures and pair nodes.
   It also contains the list of active pair nodes, overlapping pair nodes,
   colliding pair nodes, and the sorted lists of bounding boxes.
*/
typedef struct collision_state_struct
{
    int         max_polytopes;
    __col_Polytope *polytopes;
    int         num_polytopes;

#ifdef COL_DENSE_MATRIX
    __col_PairNode    *pairs;
#endif /* COL_DENSE_MATRIX */
    
#ifdef COL_SPARSE_MATRIX
    __col_PairNode   **pairs;
#endif /*COL_SPARSE_MATRIX */
    
    int          matrix_rows;
    int          matrix_cols;

    char        *overlap_status;

    __col_PairNode    *active_list;
    __col_BoxVert     *box_sort_list[3];
    __col_PairNode    *nbody_active_list;
    __col_PairNode    *colliding_list;
    char               using_nbody;
} __col_State;

/*---------------------------- Function Prototypes --------------------------*/


/*---------------------------Globals (externed)------------------------------*/



/* Protection from multiple includes. */
#endif /* INCLUDED_COLLISION_TYPES_H */



/*****************************************************************************\
\*****************************************************************************/

