#include <stdio.h>
#include "lcollision.h"
#include "ode.h"

#define EPSILON .5

static CONTACT_NODE_P contact_list = NULL;
static int check_all = 1;

int make_contact_list( OBJ_NODE_P objs ) {
    CONTACT_NODE_P pair_list = NULL, current_pair = NULL;
    OBJ_NODE_P outer_head, inner_head, current_obj;
    register int i, j;
    
    outer_head = objs;
    if( outer_head == NULL ) {
        return 0;
    }

    while( outer_head->next != NULL ) {
        inner_head = outer_head->next;
        while( inner_head != NULL ) {
            if( current_pair == NULL ) {
                contact_list = (CONTACT_NODE*)malloc( sizeof( CONTACT_NODE ) );
                contact_list->oA = &outer_head->obj;
                contact_list->oB = &inner_head->obj;
                contact_list->next = NULL;
                current_pair = contact_list;
                printf ("made contact node\n");
            } else {
                current_pair->next =
                    (CONTACT_NODE*)malloc( sizeof( CONTACT_NODE ) );
                current_pair = current_pair->next;
                current_pair->oA = &outer_head->obj;
                current_pair->oB = &inner_head->obj;
                current_pair->next = NULL;
                printf ("made contact node\n");
            }
            inner_head = inner_head->next;
        }
        outer_head = outer_head->next;
    }

    return 1;
}

int contained( OBJECT_P a, VECTOR_P point ) {
    register int i;
    POLYGON_P p;
    VECTOR_P u;
    VECTOR temp_v;

    for( i = 0; i < a->num_polygons; i++ ) {
        p = &a->polygons[i];
        u = &p->verts_world[ VERTIDX( p->index_vert[0] ) ];
        vec3sub( temp_v, point, u );
        if( vec3dot( temp_v, p->normal ) > 0.0 ) {
            return 0;
        }
    }
    return 1;
}

int collide_vert( OBJECT_P a, OBJECT_P b, VECTOR_P c_point, VECTOR_P c_dir,
    double *dist ) {
    register int i, j;
    POLYGON_P face;
    VECTOR vec;
    VECTOR_P v;
    double test_dist;
  
    for( i = 0; i < b->num_vertices; i++ ) {
        v = &b->verts_world[ VERTIDX( i ) ];
        if( contained( a, v ) ) {
            /* we got one!  find an approx collision point... */
            vec3cpy( c_point, v );
            /* find closest normal */
            *dist = 9e30;  /* far far away */
            for( j = 0; j < a->num_polygons; j++ ) {
                face = &a->polygons[j];
                vec3sub(vec, c_point, &face->verts_world[face->index_vert[0]]);
                test_dist = ABS(vec3dot( vec, face->normal ));
                if( test_dist < *dist ) {
                    *dist = test_dist; 
                    /* normal of face in a */
                    vec3cpy( c_dir, face->normal );
                }
            }
//            printf( "vertex %d of %s with value [%f,%f,%f] in obj %s.\n", i,
//                b->name, c_point[0], c_point[1], c_point[2], a->name );
            return 1;
        }
    }
    return 0;
}

int comp_double( double *i, double *j ) {
    return ( ( *i < *j ) ? -1 : 1 );
}

int collide_edge( OBJECT_P a, OBJECT_P b, VECTOR_P c_point, VECTOR_P c_dir,
    double *dist ) {
    double *isect_points;
    int num_points, i, j;
    VECTOR vec_temp, edge1, edge2, vec, proj, distv;
    VECTOR_P v1, v2, vertex;
    double d1, d2, t, test_dist, temp;
    POLYGON_P current_face;
 
    /* max number of intersection points is number of faces in A */
    isect_points = (double*)malloc( a->num_polygons * sizeof(double) );
    isect_points[0] = 0.0;
    isect_points[1] = 1.0;
    num_points = 2;
 
    for( i = 0; i < b->num_edges; i++ ) {
        v1 = &b->verts_world[ VERTIDX( b->index_edges[i*2] ) ];
        v2 = &b->verts_world[ VERTIDX( b->index_edges[i*2+1] ) ];
        for( j = 0; j < a->num_polygons; j++ ) {
            current_face = &a->polygons[j];
            vertex = &current_face->verts_world[
                VERTIDX( current_face->index_vert[0] ) ];
            vec3sub( vec_temp, v1, vertex );
            d1 = vec3dot( current_face->normal, vec_temp ); 
            vec3sub( vec_temp, v2, vertex );
            d2 = vec3dot( current_face->normal, vec_temp );
 
            if( ((d1 > 0.0) && (d2 < 1.0)) ||
                ((d1 < 0.0) && (d2 > 1.0)) ) {
                 t = ABS(d1) / ( ABS(d1) + ABS(d2) );
                 if( ( t > 0.0 ) && ( t < 1.0 ) ) {
                     isect_points[num_points] = t;
                     num_points++;
                 }
            }
        }
        qsort( isect_points, num_points, sizeof( double ), comp_double );
        for( j = 0; j < (num_points-1); j++ ) {
            t = ( isect_points[j] + isect_points[j+1] ) / 2.0;
            vec3sub( edge1, v2, v1 );
            vec3smul( vec_temp, t, edge1 );
            vec3add( vec_temp, v1, vec_temp );  /* safe */
            if( contained( a, vec_temp ) ) {
                *dist = 9e30;  /* far far away */
                for( j = 0; j < a->num_edges; j++ ) {
                    vec3sub( edge2,
                        &a->verts_world[ VERTIDX( a->index_edges[j*2] ) ],
                        &a->verts_world[ VERTIDX( a->index_edges[j*2+1] ) ] );
                    vec3sub( vec,
                        vec_temp,
                        &a->verts_world[ VERTIDX( a->index_edges[j*2+1] ) ] );
                    temp = vec3dot( edge2, vec ) / vec3dot( edge2, edge2 );
                    vec3smul( proj, temp, edge2 );
                    vec3sub( distv, vec, proj );
                    test_dist = vec3mag( distv );
                    if( test_dist < *dist ) {
                        *dist = test_dist;
                        vec3cross( vec, edge1, edge2 );
                        /* make sure pointing in right direction */
                        vec3sub( distv,
                            &a->verts_world[ VERTIDX( a->index_edges[j*2] ) ],
                            a->x );
                        if( vec3dot( vec, distv ) < 0.0 ) {
                            vec3neg( vec );
                        }
                        normalize( vec );
                        vec3cpy( c_dir, vec );
                    }
                }
                printf( "edge collision on edge ([%f,%f,%f]-[%f,%f,%f])\n   "
                    " at point [%f,%f,%f]\n", v1[0], v1[1], v1[2], v2[0], v2[1],
                    v2[2], vec_temp[0], vec_temp[1], vec_temp[2] );
                vec3cpy( c_point, vec_temp );
                free( isect_points );
                return 1;
            }
        }
    }
    free( isect_points );
    return 0;
}

int collide( OBJECT_P a, OBJECT_P b, VECTOR_P c_point, VECTOR_P c_dir,
    double *dist ) {
    if( collide_vert( b, a, c_point, c_dir, dist ) ) {
        return 1;
    }
    /*if( collide_edge( b, a, c_point, c_dir, dist ) ) {
        return 1;
    }*/
    return 0;
}

void point_velocity( VECTOR r, OBJECT_P obj, VECTOR p ) {
    VECTOR temp1, temp2;
    vec3sub( temp1, p, obj->x );
    vec3cross( temp2, obj->omega, temp1 );
    vec3add( r, obj->v, temp2 );
}

void compute_response( CONTACT_NODE_P cnode, double epsilon ) {
    VECTOR padot, pbdot, normal, ra, rb, force, vec, vec1, vec2, vec3;
    double numerator, vrel, term1, term2, term3, term4, j;
    MATRIX mat1, mat2;

    point_velocity( padot, cnode->oA, cnode->p );
    point_velocity( pbdot, cnode->oB, cnode->p );
    vec3cpy( normal, cnode->normal );
    vec3sub( ra, cnode->p, cnode->oA->x );
    vec3sub( rb, cnode->p, cnode->oB->x );

    vec3sub( vec, padot, pbdot );
    vrel = vec3dot( normal, vec );
    numerator = -( 1 + epsilon) * vrel;

    term1 = cnode->oA->mass_inv;
    term2 = cnode->oB->mass_inv;

    vec4init( vec1 ); vec4init( vec2 ); vec4init( vec3 );
    vec3cross( vec1, ra, normal );
    mat4vec4mul( vec2, cnode->oA->i_inv, vec1 );
    vec3cross( vec3, vec2, ra );
    term3 = vec3dot( normal, vec3 );

    vec4init( vec1 ); vec4init( vec2 ); vec4init( vec3 );
    vec3cross( vec1, rb, normal );
    mat4vec4mul( vec2, cnode->oB->i_inv, vec1 );
    vec3cross( vec3, vec2, rb );
    term4 = vec3dot( normal, vec3 );

    j = numerator / ( term1 + term2 + term3 + term4 );
    vec3smul( force, j, normal );

    /* apply impulse to bodies */
/*    vec3add( cnode->oA->P, cnode->oA->P, force );
    vec3sub( cnode->oB->P, cnode->oB->P, force );
    vec3cross( vec1, ra, force );
    vec3add( cnode->oA->L, cnode->oA->L, vec1 );
    vec3cross( vec1, rb, force );
    vec3sub( cnode->oB->L, cnode->oB->L, vec1 ); */
    
    vec3add( cnode->oA->force, cnode->oA->force, force );
    vec3sub( cnode->oB->force, cnode->oB->force, force );
    vec3cross( vec1, ra, force );
    vec3add( cnode->oA->torque, cnode->oA->torque, vec1 );
    vec3cross( vec1, rb, force );
    vec3sub( cnode->oB->torque, cnode->oB->torque, vec1 ); 

    /* recompute auxiliary variables */
    vec3smul( cnode->oA->v, cnode->oA->mass_inv, cnode->oA->P );
    vec3smul( cnode->oB->v, cnode->oB->mass_inv, cnode->oB->P );

    mat4vec4mul( cnode->oA->omega, cnode->oA->i_inv, cnode->oA->L );
    mat4vec4mul( cnode->oB->omega, cnode->oB->i_inv, cnode->oB->L );
}

void check_all_on() {
    CONTACT_NODE *current;
    current = contact_list;
    while( current != NULL ) {
        current->ignore = 0;
        current = current->next;
    }
}

int find_collisions() {
    CONTACT_NODE *current;
    VECTOR c_point, c_dir;
    double dist;

    current = contact_list;
    while( current != NULL ) {
        vec4init( c_point );
        vec4init( c_dir );
        if( current->ignore == 0 ) {
            if( collide( current->oA, current->oB, c_point, c_dir, &dist ) ) {
                vec3cpy( current->p, c_point );
                vec3cpy( current->normal, c_dir );
                compute_response( current, 0.5 );
//                if( ! ( dist < EPSILON ) ) {
                    // printf ("Dist: %f\n", dist); 
                    restart_ode();
//                    return 0;
//                } else {
//                    current->ignore = 1;
//                    return -1;
//                }
            }
        }
        current = current->next;
    }
    return 0;
}
