#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fparse.h"
#include "polygon.h"
#include "objmgr.h"

static int objcount = 0;

void f_errexit (char *emsg) {
  int lnum;

  lnum = getparselnum();
  fprintf (stderr, "In line %d: %s\n", lnum, emsg);
  exit (1);
}

static float vparms[9] = {0, 0, -25, 0, 0, 0, 0, 1.0, 0.0};

typedef struct INT_NODE {
    int num1;
    int num2;
    struct INT_NODE *next;
} INT_NODE, *INT_NODE_P;

/*****************************************************************************
void find_verts_polyhedron( POLYHEDRON_P p ) {
    register int i, j;
    INT_NODE_P index_list = NULL, current_index, last_index;
    int num_indices = 0, not_found;
    POLYGON_P current_poly;

    for( i = 0; i < p->num_polygons; i++ ) {
        current_poly = &p->polygons[i];
        for( j = 0; j < current_poly->num_vertices; j++ ) {
            not_found = 1;
            for( current_index = index_list; current_index != NULL;
                current_index = current_index->next ) {
                if( current_poly->index_vert[j] == current_index->num1 ) {
                    not_found = 0;
                    break;
                }
            }
            if( not_found == 1 ) {
                if( index_list == NULL ) {
                    index_list = (INT_NODE_P)malloc( sizeof( INT_NODE ) );
                    index_list->num1 = current_poly->index_vert[j];
                    index_list->next = NULL;
                    last_index = index_list;
                } else {
                    last_index->next = (INT_NODE_P)malloc( sizeof( INT_NODE ) );
                    last_index = last_index->next;
                    last_index->num1 = current_poly->index_vert[j];
                    last_index->next = NULL;
                } 
                num_indices++;
            }
        }
    }
    p->index_vert = (int*)malloc( num_indices * sizeof( int ) );
    p->num_vertices = num_indices;

    printf( "verts per poly = " );
    i = 0;
    current_index = index_list;
    while( current_index != NULL ) {
        p->index_vert[i] = current_index->num1;
        printf( "%d ", p->index_vert[i] );
        i++;
        last_index = current_index;
        current_index = current_index->next;
        free( last_index );
    }
    printf( "\n" );
}
*****************************************************************************/

#define NEXT_VERT( x, n ) ( ( ( x + 1 ) < n ) ? ( x ) + 1 : 0 )

void find_edges_polyhedron( OBJECT_P p ) {
    register int i, j;
    INT_NODE_P index_list = NULL, current_index, last_index;
    int num_indices = 0, num1, num2, temp, not_found;
    POLYGON_P current_poly;

    for( i = 0; i < p->num_polygons; i++ ) {
        current_poly = &p->polygons[i];
        for( j = 0; j < current_poly->num_vertices; j++ ) {
            num1 = current_poly->index_vert[j];
            num2 = current_poly->index_vert[NEXT_VERT( j,
                current_poly->num_vertices )];
            if( num1 > num2 ) {
                temp = num1;
                num1 = num2;
                num2 = temp;
            }
            not_found = 1;
            for( current_index = index_list; current_index != NULL;
                current_index = current_index->next ) {
                if( ( num1 == current_index->num1 ) && ( num2 ==
                    current_index->num2 ) ) {
                    not_found = 0;
                    break;
                }
            }
            if( not_found == 1 ) {
                if( index_list == NULL ) {
                    index_list = (INT_NODE_P)malloc( sizeof( INT_NODE ) );
                    index_list->num1 = num1;
                    index_list->num2 = num2;
                    index_list->next = NULL;
                    last_index = index_list;
                } else {
                    last_index->next = (INT_NODE_P)malloc( sizeof( INT_NODE ) );
                    last_index = last_index->next;
                    last_index->num1 = num1;
                    last_index->num2 = num2;
                    last_index->next = NULL;
                } 
                num_indices++;
            }
        }
    }
    p->index_edges = (int*)malloc( 2 * num_indices * sizeof( int ) );
    p->num_edges = num_indices;

    printf( "edges per poly = " );
    i = 0;
    current_index = index_list;
    while( current_index != NULL ) {
        p->index_edges[2*i] = current_index->num1;
        p->index_edges[2*i+1] = current_index->num2;
        printf( "[%d,%d] ", p->index_edges[2*i], p->index_edges[2*i+1] );
        i++;
        last_index = current_index;
        current_index = current_index->next;
        free( last_index );
    }
    printf( "\n" );
}


void f_begin (OBJECT *o) {
/* 
 * Recursively handles a begin..end block. 
 * Special handling is done on a per-object basis.
 */

  char *t, **p;
  int i, j, k, l, m, token;

  VECTOR v, u;
  FORCE_NODE *tmp;
  OBJECT *new;
  double x;

  i = gettoken(&t);
  if ((i == TOKEN_BEGIN) || (i == TOKEN_END) || 
      (i == TOKEN_STRING) || (i==TOKEN_NUMERIC) || (i==TOKEN_TEOF)) {
      f_errexit ("Reserved type expected after begin\n");  
  }
  token = i;
  switch (token) {    

    case TOKEN_OBJECT:
      /* get name */
      i = gettoken(&t);
      if (i != TOKEN_STRING) {
        f_errexit ("String type expected after object declaration\n");       
      }
      new = (OBJECT *) malloc (sizeof (OBJECT));
      memset (new, 0, sizeof(OBJECT));
      new->name = (char *) malloc (strlen(t)); 
      strcpy (new->name, t);
      new->id = objcount;
      objcount++;
      break;  
    case TOKEN_VERTEXLIST:
     /* do we have a place to stick this stuff? */
      if (o == (OBJECT *) NULL) 
        f_errexit ("Vertex list defined outside of Object context\n");
      i = gettoken(&t);
      if (i != TOKEN_NUMERIC) {
        f_errexit ("Numeric type expected after vertex declaration\n");
      }
      i = atoi (t);
      o->num_vertices = i;
      o->vertices = (VECTOR_P) malloc (i*sizeof(VECTOR));
      o->verts_world = (VECTOR_P) malloc (i*sizeof(VECTOR));
      for (j=0; j<i; j++) {
        for (k=0; k<3; k++) { 
          l = gettoken(&t);
          if (l != TOKEN_NUMERIC) {
            f_errexit ("Numeric expected\n");
          }
          o->vertices[4*j+k] = atof(t);
        }
        o->vertices[4*j+3] = 1;
      } 
      break;
    case TOKEN_FACELIST:
      if (o == (OBJECT *) NULL) 
        f_errexit ("Face list defined outside of Object context\n");
      i = gettoken(&t);
      if (i != TOKEN_NUMERIC) {
        f_errexit ("Numeric type expected after face declaration\n");
      }
      i = atoi (t);
      o->num_polygons = i; 
      o->polygons = (POLYGON *) malloc (i*sizeof(POLYGON));
      for (j=0; j<i; j++) {
        o->polygons[j].vertices = o->vertices;
        o->polygons[j].verts_world = o->verts_world; 
        k = gettoken(&t);
        if (k != TOKEN_NUMERIC) {
          f_errexit ("Numeric expected\n");
        }
        k = atoi (t);
        o->polygons[j].num_vertices = k;
        o->polygons[j].index_vert = (int *) malloc (k*sizeof(int));
        for (l=0; l<k; l++) {
          m = gettoken(&t);
          if (m != TOKEN_NUMERIC) {
            f_errexit ("Numeric expected\n");
          }
          o->polygons[j].index_vert[l] = atoi (t);
        }
        vec3sub (u, &o->vertices[o->polygons[j].index_vert[1]*4],
                 &o->vertices[o->polygons[j].index_vert[0]*4]);
        vec3sub (v, &o->vertices[o->polygons[j].index_vert[2]*4],
                 &o->vertices[o->polygons[j].index_vert[1]*4]);
        vec3cross (o->polygons[j].normal, u, v);
        normalize (o->polygons[j].normal);
      }
      break;
    case TOKEN_FORCELIST:
      if (o == (OBJECT *) NULL) {
        f_errexit ("Force list defined outside of Object context\n");
      }
      i = gettoken(&t); 
      if (i != TOKEN_NUMERIC) {
        f_errexit ("Numeric expected\n");
      }
      i = atoi(t);
      tmp = &(o->forces);
      while (tmp->next) tmp = tmp->next; 
      for (j=0; j<i; j++) {
        tmp->next = malloc (sizeof (FORCE_NODE));
        tmp = tmp->next;
        tmp->next = (FORCE_NODE *) NULL;
        for (k=0; k<3; k++) {
          l = gettoken(&t);
          if (l != TOKEN_NUMERIC) {
            f_errexit ("Numeric expected\n");
          }
          tmp->force[k] = atof(t);
        }
        tmp->force[3] = 1;
        for (k=0; k<3; k++) {
          m = gettoken(&t);
          if (m != TOKEN_NUMERIC) {
            f_errexit ("Numeric expected\n");
          }
          tmp->loc[k] = atof(t);
        }
        tmp->loc[3]=1;
      }
      break;
    case TOKEN_FLAGS:
      if (o == (OBJECT *) NULL) {
        f_errexit ("Flags defined outside of Object context\n");
      }
      i = gettoken(&t); 
      if (i != TOKEN_NUMERIC) {
        f_errexit ("Numeric expected\n");
      }
      i = atoi(t);
      for (j=0; j<i; j++) {
        k = gettoken(&t);
        if (k != TOKEN_STRING) {
          f_errexit ("String expected\n");
        }
        if (!strcmp ("stationary", t)) { 
          k = gettoken(&t);
          if (k != TOKEN_NUMERIC) {
            f_errexit ("Numeric expected\n");
          }
          o->stationary = atoi (t);
        } 
        if (!strcmp ("immobile", t)) { 
          k = gettoken(&t);
          if (k != TOKEN_NUMERIC) {
            f_errexit ("Numeric expected\n");
          }
          o->immobile = atoi (t);
        } 
      }
      break;      
    case TOKEN_TRANSLATE:
      if (o == (OBJECT *) NULL) {
        f_errexit ("Translate defined outside of Object context\n");
      }
      for (i=0; i<3; i++) {
        j = gettoken(&t);
        if (j != TOKEN_NUMERIC) {
          f_errexit ("Numeric expected\n");
        }
        o->x[i] = atof(t);
      }
      break;
    case TOKEN_ROTATE:
      if (o == (OBJECT *) NULL) {
        f_errexit ("Rotate defined outside of Object context\n");
      }
      mIdent (o->R,4);
      for (i=0; i<3; i++) {
        for (j=0; j<3; j++) {
          k = gettoken(&t);
          if (k != TOKEN_NUMERIC) {
            f_errexit ("Numeric expected\n");
          }
          o->R[i*4+j] = atof(t);
        }
      } 
      break;
    case TOKEN_VIEWPARMS:
      for (j=0; j<9; j++) {
        k = gettoken(&t);
        if (k != TOKEN_NUMERIC) {
            f_errexit ("Numeric expected\n");
        }
        vparms[j] = atof(t);
      } 
      break; 
    case TOKEN_INVMASS:
      if (o == (OBJECT *) NULL) {
        f_errexit ("Mass defined outside of Object context\n");
      }
      k = gettoken(&t);
      if (k != TOKEN_NUMERIC) {
        f_errexit ("Numeric expected\n");
      }
      o->mass_inv = atof(t);
      break;
    case TOKEN_INVTENSOR:
      if (o == (OBJECT *) NULL) {
        f_errexit ("Tensor defined outside of Object context\n");
      }
      mIdent (o->ibody_inv,4);
      for (i=0; i<3; i++) {
        for (j=0; j<3; j++) {
          k = gettoken(&t);
          if (k != TOKEN_NUMERIC) {
            f_errexit ("Numeric expected\n");
          }
          o->ibody_inv[i*4+j] = atof(t);
        }
      } 
      break;
  }
  /* handle any other nested things */
  while ((i=gettoken(&t)) != TOKEN_END) {
    if (i != TOKEN_BEGIN) {
      f_errexit ("End expected\n");
    } 
    else {
      if (token == TOKEN_OBJECT) 
        f_begin(new);
      else 
        f_begin(o);
    }
  } 

  if (token == TOKEN_OBJECT) {
    /* find_verts_polyhedron (new); */
/*    find_edges_polyhedron (new); */
    AddList (new); 
    free (new);
  }
}

int read_object_list (char *filename, float *vp) {

  char *t;
  int i;

  if (!initparse (filename)) {
    fprintf (stderr, "Cannot open %s\n", filename);
    return (-1);
  }

  while ((i = gettoken(&t)) != TOKEN_TEOF) {
    if (i != TOKEN_BEGIN) 
      f_errexit ("Begin expected\n");
    f_begin((OBJECT *)NULL); 
  }
  memcpy (vp, vparms, 9*sizeof(float)); 
  return (0);
}
