
/* File parser */
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include "fparse.h"

static char  hbuf[BUFSIZ+2];    /* The file buffer               */
static int   hindex;            /* current index into the buffer */
static int   sizehbuf;          /* size of the buffer            */
static FILE *hfile;             /* the file itself               */
static int line_no;             /* current line number           */

char hgetc (void) {
  if (hindex < sizehbuf) {
    return (hbuf[hindex++]);
  }
  hindex = 0;
  sizehbuf = fread (hbuf,1,BUFSIZ,hfile);
  hbuf[sizehbuf] = 0;
  return (hbuf[hindex++]);
}

int heof (void) {
  return (sizehbuf==0);
}

int hinit (char *filename) {
  if ((hfile = fopen (filename, "r"))==NULL)
    return 0;
  hindex = 0;
  sizehbuf = fread (hbuf,1,BUFSIZ,hfile);
  hbuf[sizehbuf] = 0;
  return 1;
}

static char tmp[BUFSIZ];

/* Returns the next token and a string containing the token */
int gettoken (char **tok) {

  int found_token=0;
  char ch; 
  int in_token=0; 
  char *t;
  int len_t=0; 
  int i;
  int flush=0;

/* tags we are parsing for, in order of TOKENS enum */
  #define NUMTAGS 15
  char *tags[] = {"/*", "*/", "begin", "end", "object","vertexlist",
                   "facelist","face","forcelist","flags", 
                   "translate", "rotate", "invmass", "invtensor", 
                   "viewparms"};
                   
   
  t = tmp; 

  while (!found_token && !heof() && len_t < BUFSIZ) {
    ch = hgetc(); 
    switch (ch) {
      case '\n':
        line_no++;
    /* Parse out whitespace */
      case '\r':
      case '\t':
      case ' ' :
        if (in_token) {
          found_token = 1;
        }
        if ((len_t > 0) && *(t-1) == '*') *(t-1) = 'A';
        break;
      case '/':
        if (len_t > 0 && *(t-1) == '*') {
        /*  
         * outside comment.  
         */
          flush = 0;
          len_t = 0;
          t = tmp;
          in_token = 0;
        }
        else {
          if (!in_token && !flush) {
            in_token = 1;
          } 
          *t++ = ch;   
          len_t++;
        }
        break;
      case '*':
        if ((len_t > 0) && *(t-1) == '/') {

        /*  
         * inside comment. start dropping stuff until we 
         * get an end comment.
         */
   
          flush = 1;
          len_t = 0;
          t = tmp;
          in_token = 0;
        }
        else  {
          if (!in_token && !flush) {
            in_token = 1;
          }
          *t++ = ch;
          len_t++;
        }
        break;
      default:
        if (!flush) {
          in_token=1;
          *t++ = ch;
          len_t++;
        } else {
        /* hack to make sure comments work */
          if ((len_t > 0) && *(t-1) == '*') *(t-1) = 'A';
        }
    } 
  }
 
  *t++ = 0;
  *tok = tmp;
    
  if (len_t == BUFSIZ) 
    return (TOKEN_OVER);
  if (heof())   
    return (TOKEN_TEOF);
 
    
  for (i=0; i<NUMTAGS; i++) {
    if (!strcmp (tmp, tags[i])) {
      return (i);
    }  
  }

 /* Either a number or a string.  Find out which. */

  for (i=0; i<len_t; i++) {
    if (!isdigit(tmp[i]) && tmp[i] != '-' && tmp[i] != '.'
        && tmp[i] !='e' )
      return (TOKEN_STRING); 
  }
  return (TOKEN_NUMERIC);
}

int initparse (char *filename) {
  line_no = 0;
  return (hinit(filename));
}

int getparselnum(void) {
  return (line_no);
}
