/*
 * GL component that renders every frame of the animation
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glut.h>
#include "polygon.h"
#include "objmgr.h"

void clear() {
    glutSwapBuffers();
    glClearColor (0.0, 0.0, 0.0, 0.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 
    glFlush();
}

void draw_scene() {
    int i, j, k, size;
    GLfloat x, y, z;
    OBJ_NODE_P current_node;
    OBJECT_P current_obj;
    POLYGON_P current_polygon;
    VECTOR_P current_vertex;

    GLfloat colors[] = { .5, 0, 0,
                         0, .5, 0,
                         0, 0, .5,
                        .5, 0, .5,
                         0, .5, .5,
                         .5, .5, .5,
                         .7, 0, 0,
                         0, .7, 0,
                         0, 0, .7,
                        .7, 0, .7,
                         0, .7, .7,
                         0, .8, .8
                        }; 

    size = SizeList();
    if( size <= 0 ) {
        return;
    }

    clear();
    for( i = 0; i < size; i++ ) {
        current_obj = IndexList( i );
        for( j = 0; j < current_obj->num_polygons; j++ ) {
            current_polygon = &current_obj->polygons[ j ];

            glBegin (GL_POLYGON);
            glColor3fv(&colors[3*((i+j)%11)]);
            for( k = 0; k < current_polygon->num_vertices; k++ ) {
                current_vertex = &current_polygon->verts_world[
                    VERTIDX( current_polygon->index_vert[ k ] ) ]; 
                x = (GLfloat) current_vertex[0];
                y = (GLfloat) current_vertex[1];
                z = (GLfloat) current_vertex[2];
                glVertex3f( x, y, z );
            }
            glEnd();
        }
    }
}

void initGLComponent( int argc, char **argv,
		      void (*simfn)(void),
		      float *eye, float *center, float *up,
		      int sizex, int sizey ) {

    glutInit( &argc, argv );  
 
    glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
    glutInitWindowSize( sizex, sizey );

    if( glutCreateWindow( "simulator" ) == GL_FALSE ) {
	printf( "error\n" );
	exit(-1);
    }

    glutDisplayFunc( draw_scene );
    glutIdleFunc( simfn );

    glClearColor( 0.0, 0.0, 0.0, 0.0 );
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glShadeModel( GL_FLAT );
    glEnable( GL_DEPTH_TEST );
    glEnable( GL_CULL_FACE );

    glMatrixMode (GL_PROJECTION);
    gluPerspective (45, 1, .01, 200);

    glMatrixMode (GL_MODELVIEW);
    gluLookAt( eye[0], eye[1], eye[2], 
	       center[0], center[1], center[2], 
	       up[0], up[1], up[2] );

    glutMainLoop();
}
