#include <stdio.h>
#include <stdlib.h>
#include <collision.h>
#include "polygon.h"
#include "objmgr.h"
#include "ode.h"

static int total_collisions = 0;
static int num_polytopes;
col_Report *all_collisions;

/* check_collisions */
/* get_collisions */
/* compute_forces */

/* make_collision_lib */
void make_collision_lib( OBJ_NODE *head_obj_node ) {
    char tempfile[255];
    int i, j;
    FILE *file;
    OBJ_NODE *current =  head_obj_node;
    OBJECT *obj;
    VECTOR_P vert;

    tmpnam( tempfile );
    file = fopen( tempfile, "a" );
    while( current != NULL ) {
        obj = &current->obj;
        fprintf( file, "polyhedron %s\n", obj->name );
        for( i = 0; i < obj->num_vertices; i++ ) {
            vert = &obj->vertices[ VERTIDX( i ) ];
            fprintf( file, "%d  %f %f %f\n", i, vert[0], vert[1], vert[2] );
        }
        fprintf( file, "*\n" );
        for( i = 0; i < obj->num_polygons; i++ ) {
            fprintf( file, "face%d", i );
            for( j = 0; j < obj->polygons[i].num_vertices; j++ ) {
                fprintf( file, "  %d", obj->polygons[i].index_vert[j] );
            }
            fprintf( file, "\n" );
        }
        fprintf( file, "*\n" );
        current = current->next;
    }
    fclose( file );
    if( col_open( -1, tempfile, 1.0, &num_polytopes ) != COL_OK ) {
        printf( "Couldn't open collision library!\n" );
        exit( 1);
    }
    current = head_obj_node;
    while( current != NULL ) {
        col_instance_polytope( current->obj.name, &current->obj.col_id );
        current = current->next;
    }    
    col_activate_all();
    col_enable_nbody();
    all_collisions = (col_Report*)malloc( num_polytopes * num_polytopes *
        sizeof( col_Report ) );
    unlink( tempfile );
}

void mat_identity( col_Mat4 m ) {
    register int i,j;

    for (i = 0; i < 4; i++)
        for(j = 0; j < 4; j++)
            m[j][i] = (i==j) ? 1.0 : 0.0;
}

void update_translations( OBJ_NODE *obj_head_node ) {
    OBJ_NODE *current;
    col_Mat4 m;

    current = obj_head_node;
    while( current != NULL ) {
        mat_identity( m );
        m[0][3] = current->obj.x[0];
        m[1][3] = current->obj.x[1];
        m[2][3] = current->obj.x[2];

        m[0][0] = current->obj.R[0];
        m[0][1] = current->obj.R[1];
        m[0][2] = current->obj.R[2];

        m[1][0] = current->obj.R[4];
        m[1][1] = current->obj.R[5];
        m[1][2] = current->obj.R[6];

        m[2][0] = current->obj.R[8];
        m[2][1] = current->obj.R[9];
        m[2][2] = current->obj.R[10];

        col_set_transform( current->obj.col_id, m );
        current = current->next;
    }
}

#define EPS .8 

void point_velocity( VECTOR r, OBJECT_P obj, VECTOR p ) {
    VECTOR temp1, temp2;
    vec3sub( temp1, p, obj->x );
    vec3cross( temp2, obj->omega, temp1 );
    vec3add( r, obj->v, temp2 );
}

int compute_face_response( OBJECT *oA, OBJECT *oB, VECTOR normal,
    double epsilon ) {
    /* note: normals should be _inward_ here */

    VECTOR force, vec, vec1, vec2, vec3;
    double numerator, vrel, j;
    MATRIX mat1, mat2;
    int i;

    vec3sub( vec, oA->v, oB->v );
    vrel = vec3dot( normal, vec );

    /* Return if bodies are moving away */
    if( vrel > 0 ) return -1;
    numerator = -( 1 + epsilon) * vrel;

    /* ignore contribution of rotational vel, there isn't one */
    j = numerator / ( oA->mass_inv + oB->mass_inv );
    vec3smul( force, j, normal );

    /* apply impulse to bodies */
    vec3add( oA->P, oA->P, force );
    vec3sub( oB->P, oB->P, force );
  
    /* recompute auxiliary variables */
    vec3smul( oA->v, oA->mass_inv, oA->P );
    vec3smul( oB->v, oB->mass_inv, oB->P );

    /* shouldn't change, but not sure if omega is even set */
    mat4vec4mul( oA->omega, oA->i_inv, oA->L );
    mat4vec4mul( oB->omega, oB->i_inv, oB->L );
    return 0;
}

int compute_response( OBJECT *oA, OBJECT *oB, VECTOR normal, VECTOR p,
    double epsilon ) {
    VECTOR padot, pbdot, ra, rb, force, vec, vec1, vec2, vec3;
    double numerator, vrel, term1, term2, term3, term4, j;
    MATRIX mat1, mat2;
    int i;
    
    point_velocity( padot, oA, p );
    point_velocity( pbdot, oB, p );
  
    printf ("normal: %f %f %f\n", normal[0], normal[1], normal[2]);
    printf ("point: %f %f %f\n", p[0], p[1], p[2]);
    printf ("omega1: %f %f %f\n", oA->omega[0], oA->omega[1], oA->omega[2]);
    printf ("omega2: %f %f %f\n", oB->omega[0], oB->omega[1], oB->omega[2]);
    printf ("x1: %f %f %f\n", oA->x[0], oA->x[1], oA->x[2]);
    printf ("x2: %f %f %f\n", oB->x[0], oB->x[1], oB->x[2]);
    printf ("padot: %f %f %f\n", padot[0], padot[1], padot[2]);
    printf ("pbdot: %f %f %f\n", pbdot[0], pbdot[1], pbdot[2]);
    printf ("vel: %f %f %f\n", oA->v[0],  oA->v[1],  oA->v[2]);
    
    vec3sub( ra, p, oA->x );
    vec3sub( rb, p, oB->x );

    vec3sub( vec, padot, pbdot );
    vrel = vec3dot( normal, vec );

   /* Return if bodies are moving away */
    if( vrel > 0 ) return -1;
    numerator = -( 1 + epsilon) * vrel;

    term1 = oA->mass_inv;
    term2 = oB->mass_inv;

    vec4init( vec1 ); vec4init( vec2 ); vec4init( vec3 );
    vec3cross( vec1, ra, normal );
    mat4vec4mul( vec2, oA->i_inv, vec1 );
    vec3cross( vec3, vec2, ra );
    term3 = vec3dot( normal, vec3 );
    printf ("term3 : %f\n", term3);

    vec4init( vec1 ); vec4init( vec2 ); vec4init( vec3 );
    vec3cross( vec1, rb, normal );
    mat4vec4mul( vec2, oB->i_inv, vec1 );
    vec3cross( vec3, vec2, rb );
    term4 = vec3dot( normal, vec3 );
    printf ("term4 : %f\n", term4);

    j = numerator / ( term1 + term2 + term3 + term4 );
    /*printf ("j = %f/(%f+%f+%f+%f) = %f\n", numerator, term1, term2, 
            term3, term4, j);*/
    vec3smul( force, j, normal );
   
    printf ("Adding momentum: %f %f %f\n", force[0], force[1], force[2]);

    /* apply impulse to bodies */
    vec3add( oA->P, oA->P, force );
    vec3cross( vec1, ra, force );
    vec3add( oA->L, oA->L, vec1 );

    vec3neg( force );
    vec3add( oB->P, oB->P, force );
    vec3cross( vec1, rb, force );
    vec3add( oB->L, oB->L, vec1 );  
  
    /* recompute auxiliary variables */
    vec3smul( oA->v, oA->mass_inv, oA->P );
    vec3smul( oB->v, oB->mass_inv, oB->P );

    mat4vec4mul( oA->omega, oA->i_inv, oA->L );
    mat4vec4mul( oB->omega, oB->i_inv, oB->L );
    return 0;
}

void find_objs( int col_id1, int col_id2, OBJECT **o1, OBJECT **o2,
    OBJ_NODE *obj_head_node ) {
    OBJ_NODE *current;
    int found1 = 0, found2 = 0;
    current = obj_head_node;
    while( current != NULL ) {
        if( current->obj.col_id == col_id1 ) {
            found1 = 1;
            *o1 = &current->obj;
        } else if( current->obj.col_id == col_id2 ) {
            found2 = 1;
            *o2 = &current->obj;
        }
        if( found1 && found2 ) {
            return;
        }
        current = current->next;
    }
}

void isect_line_plane( VECTOR point, VECTOR q, VECTOR w, VECTOR p, VECTOR v) {
    VECTOR vec1;
    double temp1, temp2;
    vec4init( vec1 );
    vec3sub( vec1, p, q );
    temp1 = vec3dot( vec1, v );
    temp2 = vec3dot( w, v );
    vec3smul( vec1, ( temp1 / temp2 ), w );
    vec3add( point, q, vec1 );
}

int compute_collisions( OBJ_NODE *obj_head_node ) {
    int i;
    int num_collisions = 0, num_resp = 0;
    __col_Vertex *vert1, *vert2;
    __col_Face *face1, *face2;
    __col_Edge *edge1, *edge2;
    VECTOR normal, point, dir1, dir2, p1, p2, vec1, vec2, vec3, plane;
    double m[16];
    double temp1, temp2, d, t;
    OBJECT *o1, *o2;
    int tag1, tag2;

    if( col_test() != COL_OK ) {
        printf( "col_test didn't return COL_OK!\n" );
        exit( 1 );
    }
    num_collisions = col_report_collision_all( all_collisions,
        num_polytopes * num_polytopes );

    for( i = 0; i < num_collisions; i++ ) {
        /*printf( "object %d colliding with object %d.\n",
            all_collisions[i].ids[0], all_collisions[i].ids[1] );*/
        vert1 = (__col_Vertex*)all_collisions[i].features[0];
        vert2 = (__col_Vertex*)all_collisions[i].features[1];
        tag1 = vert1->tag;
        tag2 = vert2->tag;
        printf( "type of collision is %d / %d\n", tag1, tag2 );

        if( ( tag1 == 3 ) && ( tag2 == 3 ) ) {
            /* face/face collision, pick a normal */
            find_objs( all_collisions[i].ids[0], all_collisions[i].ids[1],
                &o1, &o2, obj_head_node );
            face1 = (__col_Face*)all_collisions[i].features[0];
            face2 = (__col_Face*)all_collisions[i].features[1];
#ifdef SHRUBBERY
            /* points of collision are the point at which the vector from
               center intersects the plane */
            vec4init(point); vec4init( normal );
             
            /* = intersection of line between CM's and one of the 
              plane eq's */ 

            vec4init (vec1);
            vec3sub (vec1, o2->x, o1->x);  
           
            temp2 = face2->plane[0] * vec1[0] + 
                    face2->plane[1] * vec1[1] +
                    face2->plane[2] * vec1[2];
 
            temp1 = - (face2->plane[0] * o1->x[0] + 
                       face2->plane[1] * o1->x[1] +
                       face2->plane[2] * o1->x[2] +
                       face2->plane[3]);

            if (temp2 == 0) {
             /* planes perpendicular. Real odd.  Not going to bother
                fixing for now.  */ 
              point[0] = 0;  point[1] = 0;  point[2] = 0;
            } else {
              temp1 /= (float)temp2;
              vec3smul (point, temp1, vec1);
              vec3add (point, point, o1->x);
            }
#endif
            vec4init( normal );
            /* get normal of b */
            mMul( normal, o2->R, face2->plane, 4, 4, 1 );

            if (compute_face_response( o1, o2, normal, EPS ) != -1) {
              num_resp++;
              restart_ode();
            }

        } else if( ( tag1 == 1 ) && ( tag2 == 3 ) ) {
            /* vert / face collision, pick normal of face */
            find_objs( all_collisions[i].ids[0], all_collisions[i].ids[1],
                &o1, &o2, obj_head_node );
            vert1 = (__col_Vertex*)all_collisions[i].features[0];
            face1 = (__col_Face*)all_collisions[i].features[1];
            vec4init( point ); vec4init( normal );
            point[0] = vert1->xcoords[0];
            point[1] = vert1->xcoords[1];
            point[2] = vert1->xcoords[2];

            mMul( normal, o2->R, face1->plane, 4, 4, 1 );

            if (compute_response( o1, o2, normal, point, EPS ) != -1) {
              num_resp++;
              restart_ode();
            }

        } else if( ( tag1 == 3 ) && ( tag2 == 1 ) ) {
            /* vert / face collision, pick normal of face */
            find_objs( all_collisions[i].ids[0], all_collisions[i].ids[1],
                &o1, &o2, obj_head_node );
            vert1 = (__col_Vertex*)all_collisions[i].features[1];
            face1 = (__col_Face*)all_collisions[i].features[0];
            vec4init( point ); vec4init( normal );
            point[0] = vert1->xcoords[0];
            point[1] = vert1->xcoords[1];
            point[2] = vert1->xcoords[2];
            mMul( normal, o1->R, face1->plane, 4, 4, 1 );
            vec3neg( normal );

            if (compute_response( o1, o2, normal, point, EPS ) != -1) {
              num_resp++;
              restart_ode();
            }

        } else if( ( tag1 == 2 ) && ( tag2 == 2 ) ) {
            /* edge / edge collision */
            find_objs( all_collisions[i].ids[0], all_collisions[i].ids[1],
                &o1, &o2, obj_head_node );
            edge1 = (__col_Edge*)all_collisions[i].features[0];
            edge2 = (__col_Edge*)all_collisions[i].features[1];
            vec4init( dir1 ); vec4init( dir2 );
            vec4init( p1 ); vec4init( p2 );
            vec4init( vec1 ); vec4init( vec2 );
            vec4init( normal ); vec4init( point );
            vec4init( plane );
            // damn bugs in icollide, gotta calc dirs ourselves...
            //dir1[0] = edge1->xu[0];
            //dir1[1] = edge1->xu[1];
            //dir1[2] = edge1->xu[2];
            //dir2[0] = edge2->xu[0];
            //dir2[1] = edge2->xu[1];
            //dir2[2] = edge2->xu[2];

            dir1[0] = edge1->v2->xcoords[0] - edge1->v1->xcoords[0];
            dir1[1] = edge1->v2->xcoords[1] - edge1->v1->xcoords[1];
            dir1[2] = edge1->v2->xcoords[2] - edge1->v1->xcoords[2];
            normalize( dir1 );

            dir2[0] = edge2->v2->xcoords[0] - edge2->v1->xcoords[0];
            dir2[1] = edge2->v2->xcoords[1] - edge2->v1->xcoords[1];
            dir2[2] = edge2->v2->xcoords[2] - edge2->v1->xcoords[2];
            normalize( dir2 );

            p1[0] = edge1->v1->xcoords[0];
            p1[1] = edge1->v1->xcoords[1];
            p1[2] = edge1->v1->xcoords[2];
            p2[0] = edge2->v1->xcoords[0];
            p2[1] = edge2->v1->xcoords[1];
            p2[2] = edge2->v1->xcoords[2];
            vec3cross( normal, dir2, dir1 );

            // To find intersection, create plane containing p2 and the
            // the normal (which also contains all of dir2)
            vec3cross( plane, normal, dir2 );
            d = -vec3dot( plane, p2 );
            printf( "d = %f, plane =\n", d );
            print_vec4( plane );
            printf( "dir1 =\n" );
            print_vec4( dir1 );
            printf( "dir2 =\n" );
	    print_vec4( dir2 );
            printf( "normal =\n" );
	    print_vec4( normal );
            printf( "p1 =\n" );
	    print_vec4( p1 );
            printf( "p2 =\n" );
	    print_vec4( p2 );


            // now find the intersection of the line p1+t*dir1 with that
            // plane

            // should probably be checking for underflow...
            temp1 = vec3dot( plane, p1 ) + d;
            temp2 = vec3dot( plane, dir1 );

	    if( temp2  == 0 ) {
                printf( "type... whoa, underflow!\n" );
            }
            t = temp1 / temp2;
            if( ( t > 1.0 ) || ( t < 0.0 ) ) {
                printf( "type... whoa, t is bad: %f\n", t );
            }
            point[0] = p1[0] + t * dir1[0];
            point[1] = p1[1] + t * dir1[1];
            point[2] = p1[2] + t * dir1[2];
            printf( "point =\n" );
	    print_vec4( point );

            //point[0] = .5 * (edge1->v1->xcoords[0] + edge1->v2->xcoords[0]);
            //point[1] = .5 * (edge1->v1->xcoords[1] + edge1->v2->xcoords[1]);
            //point[2] = .5 * (edge1->v1->xcoords[2] + edge1->v2->xcoords[2]);

            if (compute_response( o1, o2, normal, point, EPS ) != -1) {
              num_resp++;
              restart_ode();
            }
 
        } else if ((tag1 == 3) && (tag2 == 2))  {

            find_objs( all_collisions[i].ids[0], all_collisions[i].ids[1],
                &o1, &o2, obj_head_node );
            face1 = (__col_Face*)all_collisions[i].features[0];
            edge1 = (__col_Edge*)all_collisions[i].features[1];
            vec4init( point ); vec4init( normal );
            point[0] = .5 * (edge1->v1->xcoords[0] + edge1->v2->xcoords[0]);
            point[1] = .5 * (edge1->v1->xcoords[1] + edge1->v2->xcoords[1]);
            point[2] = .5 * (edge1->v1->xcoords[2] + edge1->v2->xcoords[2]);
            mMul( normal, o1->R, face1->plane, 4, 4, 1 );
            vec3neg( normal );

            if (compute_response( o1, o2, normal, point, EPS ) != -1) {
              num_resp++;
              restart_ode();
            }

        } else if ((tag1 == 2) && (tag2 == 3))  {

            find_objs( all_collisions[i].ids[0], all_collisions[i].ids[1],
                &o1, &o2, obj_head_node );
            edge1 = (__col_Edge*)all_collisions[i].features[0];
            face1 = (__col_Face*)all_collisions[i].features[1];
            vec4init( point ); vec4init( normal );
            point[0] = .5 * (edge1->v1->xcoords[0] + edge1->v2->xcoords[0]);
            point[1] = .5 * (edge1->v1->xcoords[1] + edge1->v2->xcoords[1]);
            point[2] = .5 * (edge1->v1->xcoords[2] + edge1->v2->xcoords[2]);
            mMul( normal, o2->R, face1->plane, 4, 4, 1 );

            if (compute_response( o1, o2, normal, point, EPS ) != -1) {
              num_resp++;
              restart_ode();
            }

        } else if ((tag1 == 2) && (tag2 == 1))  {

            find_objs( all_collisions[i].ids[0], all_collisions[i].ids[1],
                &o1, &o2, obj_head_node );
            edge1 = (__col_Edge*)all_collisions[i].features[0];
            vert1 = (__col_Vertex*)all_collisions[i].features[1];
            vec4init( point ); vec4init( normal );
            point[0] = vert1->xcoords[0];
            point[1] = vert1->xcoords[1];
            point[2] = vert1->xcoords[2];
            vec1[0] = -(edge1->fl->plane[0] + edge1->fr->plane[0]);
            vec1[1] = -(edge1->fl->plane[1] + edge1->fr->plane[1]);
            vec1[2] = -(edge1->fl->plane[2] + edge1->fr->plane[2]);
            normalize (vec1);
            mMul( normal, o1->R, vec1, 4, 4, 1 );
            vec3neg( normal );

            if (compute_response( o1, o2, normal, point, EPS ) != -1) {
              num_resp++;
              restart_ode();
            }

        } else if ((tag1 == 1) && (tag2 == 2))  {

            find_objs( all_collisions[i].ids[0], all_collisions[i].ids[1],
                &o1, &o2, obj_head_node );
            vert1 = (__col_Vertex*)all_collisions[i].features[0];
            edge1 = (__col_Edge*)all_collisions[i].features[1];
            vec4init( point ); vec4init( normal );
            point[0] = vert1->xcoords[0];
            point[1] = vert1->xcoords[1];
            point[2] = vert1->xcoords[2];
            vec1[0] = edge1->fl->plane[0] + edge1->fr->plane[0];
            vec1[1] = edge1->fl->plane[1] + edge1->fr->plane[1];
            vec1[2] = edge1->fl->plane[2] + edge1->fr->plane[2];
            normalize (vec1);
            mMul( normal, o2->R, vec1, 4, 4, 1 );

            if (compute_response( o1, o2, normal, point, EPS ) != -1) {
              num_resp++;
              restart_ode();
            }
        }
         else {
            printf( "Warning!  Got degenerate collision of type %d / %d.\n",
                tag1, tag2 );
            restart_ode();
        }
    }
    return (num_resp > 0);
}



