/*  Main program code.  Includes top level calls to simulation run.
 *
 *  $Id: main.c,v 1.12 1997/06/06 02:56:39 ant Exp len $
 */

#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include "polygon.h"
#include "fileio.h"
#include "sim.h"
#include "ode.h"
#include "objmgr.h"
#include "gltest.h"
#include "icollision.h"
#include "screencap.h"

#define ENDSIM 10          /* Length to run simulation        */
#define TIME_EPSILON 1e-4  /* The size of collision intervals */
#define ODE_EPSILON .1     /* Something for Ode solver.       */
#define TIMESTEP 1.0/30    /* Normal frame size               */

static float vp[9];
static double *y;          /* Stores integration results */
static double t = 0.0;     /* Time   p                   */

OBJ_NODE ObjList;
void initscene (char *fname) {
/*
 * void initscene (void);
 * 
 * does all the startup jazz
 *
 */

 /* Let the object mgr know what's going on */
  SetList (&ObjList);  

 /* Read in objects */
  if (read_object_list (fname, vp) == -1) exit(1); 

  make_collision_lib( ObjList.next );
}

void do_ode (double *y, int n, double t, double deltat, int level) {
  
  int ngood, nbad;

#ifdef DEBUG
  printf ("%*sdo_ode (%p, %d, %f, %f)\n", -2*level, " ", y, n, t, deltat); 
#endif

  if (ode (y, n, t, t+deltat, ODE_EPSILON, 1000, 0, 
      &ngood, &nbad, dydt) == -1 && deltat > TIME_EPSILON) {
    do_ode (y, n, t, deltat/2, level+1);
    do_ode (y, n, t+deltat/2, deltat/2, level+1);
  } else {
     OBJECT *o;
     Array_to_Bodies(y);
     o = IndexList(0);

#ifdef DEBUG
     printf ("x: %f %f\n", t, o->x[1]);
     printf ("v: %f %f\n", t, o->v[1]);
     printf ("f: %f %f\n", t, o->force[1]);
     printf ("P: %f %f\n", t, o->P[1]);
#endif

  }
}   

void next_sim_step() {

  OBJECT *o;
  int i,j,k;
  int num_objs;

#ifdef CAPTURE
  capScreen();                         // save old image
#endif

  if ( t > ENDSIM ) 
    exit(1);

  num_objs = SizeList();

  /* Calculate new forces, torques, etc. */ 
  do_ode (y, STATE_SIZE*num_objs, t, TIMESTEP,0);

  /* Update WC for all objects */
  for (k=0; k<num_objs; k++) {
    o = IndexList(k);

    for ( i=0; i<o->num_vertices; i++ ) {
      mat4vec4mul (&(o->verts_world[i*4]), o->R, &(o->vertices[i*4]));
      for (j=0; j<3; j++) {
         o->verts_world[i*4+j] = o->verts_world[i*4+j] + o->x[j];
      }
    }
  }

  t += TIMESTEP;
  glutPostRedisplay();
} 

int main (char argc, char *argv[]) {

  int num_objs;

  if (argc<1) {
    printf ("Usage: %s <file>\n", argv[0]);
  }
  initscene (argv[1]);

#ifdef CAPTURE
  initScreenCap( "dump/sim", 400, 400 );
#endif

  num_objs = SizeList();
  y = (double *) malloc (sizeof (double) * num_objs * STATE_SIZE);

  InitSim();
  Bodies_to_Array(y);

  initGLComponent( argc, argv, next_sim_step,
                   &vp[0], &vp[3], &vp[6],
                   400, 400 );
  destroyScreenCap();
}

