#include <stdio.h>
#include "polygon.h"
#include "math.h"

double vec3mag( VECTOR u ) {
    return sqrt( u[0] * u[0] + u[1] * u[1] + u[2] * u[2] );
}

void normalize( VECTOR u ) {
    double mag;
    mag = vec3mag( u );
    u[0] = u[0] / mag;
    u[1] = u[1] / mag;
    u[2] = u[2] / mag;
    u[3] = 1.0;
}

void vec3init( VECTOR u ) {
    u[0] = u[1] = u[2] = 0.0;
}

void vec4init( VECTOR u ) {
    u[0] = u[1] = u[2] = 0.0;
    u[3] = 1.0;
}

void vec3cpy( VECTOR u, VECTOR v ) {
    u[0] = v[0];
    u[1] = v[1];
    u[2] = v[2];
}

void vec4cpy( VECTOR u, VECTOR v ) {
    u[0] = v[0];
    u[1] = v[1];
    u[2] = v[2];
    u[3] = v[3];
}

void vec3add( VECTOR r, VECTOR u, VECTOR v ) {
    r[0] = u[0] + v[0];
    r[1] = u[1] + v[1];
    r[2] = u[2] + v[2];
}

void vec3sub( VECTOR r, VECTOR u, VECTOR v ) {
    r[0] = u[0] - v[0];
    r[1] = u[1] - v[1];
    r[2] = u[2] - v[2];
}

void vec3smul( VECTOR r, double s, VECTOR v ) {
    r[0] = s * v[0];
    r[1] = s * v[1];
    r[2] = s * v[2];
}

double vec3dot( VECTOR u, VECTOR v ) {
   return ( u[0] * v[0] + u[1] * v[1] + u[2] * v[2] );
}

void vec3cross( VECTOR r, VECTOR u, VECTOR v ) {
    r[0] = u[1] * v[2] - u[2] * v[1];
    r[1] = u[2] * v[0] - u[0] * v[2];
    r[2] = u[0] * v[1] - u[1] * v[0];
}

void vec3neg( VECTOR u ) {
    u[0] = -u[0]; u[1] = -u[1]; u[2] = -u[2];
}

void mat4mul( MATRIX r, MATRIX m, MATRIX n ) {
    register int i, j;
    for( i = 0; i < 4; i++ ) {
        for( j = 0; j < 4; j++ ) {
            r[i][j] = m[i][0] * n[0][j] +
                      m[i][1] * n[1][j] +
                      m[i][2] * n[2][j] +
                      m[i][3] * n[3][j];
        }
    }
}

void mat4vec4mul( VECTOR v, double *m, VECTOR u ) {
    register int i;
    for( i = 0; i < 4; i++ ) {
        v[i] = m[4*i+0] * u[0] +
               m[4*i+1] * u[1] +
               m[4*i+2] * u[2] +
               m[4*i+3] * u[3];
    }
}

void idmat4( MATRIX m ) {
    m[0][1] = m[0][2] = m[0][3] = 0;
    m[1][0] = m[1][2] = m[1][3] = 0;
    m[2][0] = m[2][1] = m[2][3] = 0;
    m[3][0] = m[3][1] = m[3][2] = 0;
    m[0][0] = m[1][1] = m[2][2] = m[3][3] = 1;
}

void print_vec4( VECTOR v ) {
    int i;
    for( i = 0; i < 4; i++ ) {
        printf( "[% 10g ]\n", v[i] );
    }
}

void print_mat4( MATRIX m ) {
    int row, col;
    for( row = 0; row < 4; row++ ) {
        printf( "[" );
        for( col = 0; col < 4; col++ ) {
            printf( "% 10g", m[row][col] );
        }
        printf( " ]\n" );
    }
}
