#ifndef PPOLY
#define PPOLY
typedef double VECTOR[4];
typedef double *VECTOR_P;
typedef double MATRIX[4][4];
typedef double *MATRIX_P;

typedef struct _Flist {
    VECTOR force;
    VECTOR loc;
    struct _Flist *next;
} FORCE_NODE;

#define VERTIDX( i )  ( i * 4 )

typedef struct POLYGON {
    VECTOR_P vertices; /* array of vertices that this polygon is a subset of */
    VECTOR_P verts_world;
    int *index_vert;   /* malloced array of indices for this poly's vertices */
    int num_vertices;  /* number of vertices for _this_ polygon */

    VECTOR normal;     /* normal pointing outward */
} POLYGON, *POLYGON_P;

typedef struct POLYHEDRON {
    POLYGON_P polygons;   /* malloced array of polygons */
    int num_polygons;     /* number of polygons */

    VECTOR_P vertices;
    VECTOR_P verts_world;
    int *index_vert;      /* malloced array of vertices for this polyhedron */
    int num_vertices;     /* number of vertices for this polyhedron */

    int *index_edges;    /* array of indices denoting edges, paired */
    int num_edges;       /* number of edges, .5 length of array */

} POLYHEDRON, *POLYHEDRON_P;

typedef struct OBJECT {
    char *name;               /* name of object */
    int id;                   /* defined at runtime */
    int col_id;               /* defined at by collision det. library */
    int colliding;
    int old_colliding;

    VECTOR_P vertices;        /* malloced array of vertices */
    VECTOR_P verts_world;     /* same vertices, but in world coordinates */
    int num_vertices;         /* number of vertices */

    int *index_edges;         /* array of indices denoting edges, paired */
    int num_edges;            /* number of edges, .5 length of array */

    POLYGON_P polygons;       /* malloced array of polygons */
    int num_polygons;         /* number of polygons */

   /* Constants */
    double mass_inv;
    double ibody[16];
    double ibody_inv[16];

   /* State variables */
    VECTOR x;      
    VECTOR q;                
    VECTOR P, L;

   /* derived each frame */
    double i_inv[16];
    double R[16];
    VECTOR v, omega;

   /* computed */
    VECTOR force, torque;

    int stationary;          /* is it currently stable? */
    int immobile;            /* for floors and fixed objects, ignores g */
   
    FORCE_NODE forces;
} OBJECT, *OBJECT_P;

typedef struct OBJ_NODE {
    OBJECT obj;
    struct OBJ_NODE *next;
} OBJ_NODE, *OBJ_NODE_P;


#define ABS( x )  ( ( x ) > 0 ? ( x ) : -( x ) )

double vec3mag( VECTOR u );
void normalize( VECTOR u );
void vec3init( VECTOR u );
void vec4init( VECTOR u );
void vec3cpy( VECTOR u, VECTOR v );
void vec4cpy( VECTOR u, VECTOR v );
void vec3add( VECTOR r, VECTOR u, VECTOR v );
void vec3sub( VECTOR r, VECTOR u, VECTOR v );
void vec3smul( VECTOR r, double s, VECTOR v );
double vec3dot( VECTOR u, VECTOR v );
void vec3cross( VECTOR r, VECTOR u, VECTOR v );
void vec3neg( VECTOR u );
void mat4mul( MATRIX r, MATRIX m, MATRIX n );
void mat4vec4mul( VECTOR v, double *m, VECTOR u );
void idmat4( MATRIX m );
#endif
