/*
 *  This code can be used to generate screen captures, to make a movie
 *  for example.  It writes out SGI RGB files, which are eminently 
 *  useless except for the sgi makemovie program.
 */

#include <stdio.h>
#include <string.h>
#include <GL/gl.h>

#define MAGIC 474
enum encode_types {
  VERBATIM,
  RLE 
};

static char basename[ 256 ];
static int num_cap = 0;
static int cap_w = 100;
static int cap_h = 100;
static char *pixels;

void writeShortBigEndian( short c, FILE *fp );
void writeLongBigEndian( long c, FILE *fp );

int writeSGIimage( char *filename, int width, int height, 
                   char *rgbpixels ) {
  FILE *fp;
  char desc[80];
  int i, j;

  fp = fopen( filename, "wb" );
  if (!fp) 
    return -1;
  
  memset( desc, 0, 80 );
  strncpy( desc, filename, 79 );
  
  /* Write header */
  writeShortBigEndian( MAGIC, fp );   // Magic number
  fputc( VERBATIM, fp );              // encoding
  fputc( 1, fp );                     // bytes per channel
  writeShortBigEndian( 3, fp );       // dimension (x,y,components) 
  writeShortBigEndian( width, fp );   // xsize
  writeShortBigEndian( height, fp );  // ysize
  writeShortBigEndian( 3, fp );       // # channels (r,g,b) 
  writeLongBigEndian( 0, fp );        // min pixel value 
  writeLongBigEndian( 255, fp );      // max pixel value 
  writeLongBigEndian( 0, fp );        // placeholder

  fwrite( desc, 1, 80, fp );          // description
  writeLongBigEndian( 0, fp );        // colormap (flat)

  for( i=0; i<404; i++ )              // filler
    fputc( 0, fp );  

  /* write pixel data, a channel at a time */
  for( i=0; i<3; i++ ) {
    for( j=0; j<width * height; j++ ) {
      fputc( rgbpixels[ j * 3 + i ], fp );
    }  
  }

  fclose( fp );
  return 1;
}

void writeShortBigEndian( short c, FILE *fp ) {
  
  unsigned char b;
  
  b = ( c & 0xff00 ) >> 8;
  fputc( b, fp ); 

  b = ( c & 0xff );
  fputc( b, fp ); 
}

void writeLongBigEndian( long c, FILE *fp ) {

  short hw;
  hw = ( c & 0xffff0000 ) >> 16;
  writeShortBigEndian( hw, fp );

  hw = ( c & 0xffff);
  writeShortBigEndian( hw, fp );
}

void initScreenCap( char *prefix, int width, int height ) {
  strncpy( basename, prefix, 255 );   
  cap_w = width;
  cap_h = height;
  pixels = (char *) malloc( 3 * width * height );
}

void destroyScreenCap() {
  free( pixels );
}

void capScreen() {

  char filename[ 256 ];

  snprintf( filename, 256, "%s%05d.sgi", basename, num_cap );
  glReadPixels( 0, 0, cap_w, cap_h, GL_RGB, GL_UNSIGNED_BYTE, pixels );
  writeSGIimage( filename, cap_w, cap_h, pixels );  
  num_cap++; 
}
